/***** genDiv.c ***********************************
 * Description: Routines for measuring genetic 
 *   diversity.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * Date: Tue Feb 16 21:33:24 2010
 **************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include "interface.h"
#include "genDiv.h"
#include "queue.h"
#include "eprintf.h"

void findPoly(Alignment *al);
void prepareWin(Alignment *al, int winLen, int stepLen);

double td(double a1, double a2, double s, double p, int n);

void printPoly(Alignment *al){
  int i, j, c;

  if(al->poly == NULL)
    findPoly(al);
  /* Print positions of polymorphisms */
  printf(">Positions_frequencies:%ld\n",al->numPoly);
  for(i=0;i<al->numPoly;i++)
    printf("%ld\t%.3f\n",al->poly[i]+1,al->polyFreq[i]);
  /* Print alleles at polymorphic positions */
  for(i=0;i<al->m;i++){
    printf("%s\n",al->headers[i]);
    c=0;
    for(j=0;j<al->numPoly;j++){
      printf("%c",al->al[i][al->poly[j]]);
      c++;
      if(c==60){
	printf("\n");
	c = 0;
      }
    }
    if(c)
      printf("\n");
  }
}

/* pi: number of average pairwise differences per site */
double pi(Alignment *al){
  int i, j, k;
  double  s1, s2;

  if(al->poly == NULL){
    findPoly(al);
  }
  s2 = 0;
  for(i=0;i<al->m-1;i++){
    for(j=i+1;j<al->m;j++){
      s1 = 0;
      for(k=0;k<al->numPoly;k++){
	if(al->al[i][al->poly[k]] != al->al[j][al->poly[k]])
	  s1++;
      }
      s2 += s1;
    }
  }
  
  s2 /= al->m * (al->m - 1) / 2;

  return s2;
}
/* tajima: Tajima's D */
double tajima(Alignment *al){
  double s, p, a1, a2, n;
  int i;

  s = ss(al);
  p = pi(al);
  n = al->m;

  a1 = 0;
  a2 = 0;
  for(i=1;i<n;i++){
    a1 += 1.0/i;
    a2 += (1.0/i/i);
  }

  return td(a1, a2, s, p, n);
}

/* ss: number of segregating sites per site */
double ss(Alignment *al){

  if(al->poly == NULL)
    findPoly(al);

  return (double)al->numPoly;
}

double *winPi(Args *args, Alignment *al){
  int i, j, k, numWin;
  int lb, rb;
  double *arr, *pol, factor;
  double sum;
  
  arr = (double *)emalloc(al->n*sizeof(double));
  pol = (double *)emalloc(al->n*sizeof(double));
  if(al->poly == NULL)
    findPoly(al);
  if(al->polInd == NULL)
    prepareWin(al,args->w,args->S);
  /* compute average number of mismatches for all positions */
  factor = al->m*(al->m-1)/2;
  for(i=0;i<al->n;i++){
    pol[i] = 0;
    if(al->polInd[i]){
      for(j=0;j<al->m-1;j++)
	for(k=j+1;k<al->m;k++)
	  if(al->al[j][i] != al->al[k][i])
	    pol[i]++;
      pol[i] /= factor;
    }
  }
  numWin = 0;
  /* take care of first window */
  rb = 0;
  sum = 0;
  while(rb < args->w && rb < al->n)
    sum += pol[rb++];
  arr[numWin++] = sum;
  /* scan remaining windows */
  lb = 0;
  while(rb < al->n-args->S+1){
    for(i=0;i<args->S;i++){
      sum += pol[rb+i];
      sum -= pol[lb+i];
    }
    rb += args->S;
    lb += args->S;
    arr[numWin++] = sum;
  }
  free(pol);
  return arr;
}

/* winSs: sliding window analysis of segregating sites */
double *winSs(Args *args, Alignment *al){
  int i, numWin, numPol, rb, lb;
  double *arr;

  arr = (double *)emalloc(al->n*sizeof(double));
  if(al->poly == NULL)
    findPoly(al);
  if(al->polInd == NULL)
    prepareWin(al,args->w,args->S);
  numWin = 0;
  /* take care of first window */
  rb = 0;
  numPol = 0;
  while(rb < args->w && rb < al->n)
    numPol += al->polInd[rb++];
  arr[numWin++] = numPol;
  /* scan remaining windows */
  lb = 0;
  while(rb < al->n-args->S+1){
    for(i=0;i<args->S;i++){
      if(al->polInd[rb+i])
	 numPol++;
      if(al->polInd[lb+i])
	numPol--;
    }
    rb += args->S;
    lb += args->S;
    arr[numWin++] = numPol;
  }
  return arr;
}

/* td: compute Tajima's D */
double td(double a1, double a2, double s, double p, int n){
  double b1, b2, c1, c2, e1, e2;
  double d;

  b1=(n+1)/3./(n-1);
  b2=2.*(n*n+n+3.)/(9.*n*(n-1));
  c1=b1-1./a1;
  c2=b2-(n+2)/(a1*n)+a2/a1/a1;
  e1=c1/a1;
  e2=c2/(a1*a1+a2);

  if(s>0)
    d=(p-s/a1)/sqrt(e1*s+e2*s*(s-1));
  else
    d=0.;
  
  return d;
}

double *winTajima(Args *args, Alignment *al){
  long i;
  double *pArr, *sArr, *dArr;
  double a1, a2;

  a1 = 0;
  a2 = 0;
  for(i=1;i<al->m;i++){
    a1 += 1.0/i;
    a2 += (1.0/i/i);
  }

  pArr = winPi(args, al);
  sArr = winSs(args, al);
  dArr = (double *)emalloc((al->n-args->w+1)*sizeof(double));

  for(i=0;i<al->n-args->w+1;i++){
    dArr[i] = td(a1, a2, sArr[i], pArr[i], al->m);
  }
  return dArr;
}


Alignment *seq2al(Sequence *seq){
  int i;
  Alignment *al;

  al = (Alignment *)emalloc(sizeof(Alignment));
  al->m = seq->numSeq;
  al->headers = seq->headers;
  al->n = seq->borders[0];
  al->al = (char **)emalloc(al->m*sizeof(char *));
  al->al[0] = seq->seq;
  for(i=1;i<seq->numSeq;i++)
    al->al[i] = seq->seq + seq->borders[i-1] + 1;

  al->numPoly = 0;
  al->poly = NULL;
  al->polInd = NULL;
  al->nucInd = NULL;
  return al;
}

/* findPoly: fills an array of positions that are polymorphic 
*     and consist solely of the four canonical nucleotides.
*/
void findPoly(Alignment *al){
  int i, j, c, p;
  int *dic;

  dic = getRestrictedDnaDictionary(NULL);
  al->poly = (long *)emalloc(al->n*sizeof(long));
  al->nuc = (long*)emalloc(al->n*sizeof(long));
  al->polyFreq = (float *)emalloc(al->n*sizeof(float));

  al->numPoly = 0;
  al->numNuc = 0;
  for(i=0;i<al->n;i++){
    p = 0;
    if(dic[(int)al->al[0][i]])
      c = 1;
    else{
      c = 0;
      continue;
    }
    for(j=1;j<al->m;j++){
      if(!dic[(int)al->al[j][i]]){
	c = 0;
	break;
      }
      if(al->al[0][i] != al->al[j][i])
	p = 1;
    }
    if(c){
      al->nuc[al->numNuc++] = i;
      if(p){
	al->poly[al->numPoly] = i;
	al->polyFreq[al->numPoly] = 1.;
	for(j=1;j<al->m;j++)
	  if(al->al[j][i] == al->al[0][i])
	    al->polyFreq[al->numPoly]++;
	al->polyFreq[al->numPoly] /= (float)al->m;
	if(al->polyFreq[al->numPoly] > 0.5)
	  al->polyFreq[al->numPoly] = 1. - al->polyFreq[al->numPoly];
	al->numPoly++;
      }
    }
  }
  if(al->numPoly)
    al->poly = (long *)erealloc(al->poly,al->numPoly*sizeof(long));
  if(al->numNuc)
    al->nuc = (long *)erealloc(al->nuc,al->numNuc*sizeof(long));
  if(al->numPoly)
    al->polyFreq = (float *)erealloc(al->polyFreq,al->numPoly*sizeof(float));
  free(dic);
}

/* prepareWin: prepare sliding window analysis
 *   should be preceded by findPoly, though this 
 *   is checked for
 */
void prepareWin(Alignment *al, int winLen, int stepLen){
  int i, j, rb, lb, numNuc;

  if(al->poly == NULL)
    findPoly(al);
  al->polInd = (char *)emalloc(al->n*sizeof(char));
  al->nucInd = (char *)emalloc(al->n*sizeof(char));

  /* mark canonical nucleotides */
  j = 0;
  for(i=0;i<al->numNuc;i++){
    while(j<al->nuc[i])
      al->nucInd[j++] = 0;
    al->nucInd[j++] = 1;
  }
  for(i=j;i<al->n;i++)
    al->nucInd[i] = 0;

  /* mark polymorphisms */
  j = 0;
  for(i=0;i<al->numPoly;i++){
    while(j<al->poly[i])
      al->polInd[j++] = 0;
    al->polInd[j++] = 1;
  }
  for(i=j;i<al->n;i++)
    al->polInd[i] = 0;
  /* count canonical nucleotides per window */
  al->winNumNuc = (int *)emalloc(al->n*sizeof(int));
  /* take care of first window */
  rb = 0;
  numNuc = 0;
  al->numWin = 0;
  while(rb < winLen && rb < al->n)
    numNuc += al->nucInd[rb++];
  al->winNumNuc[al->numWin++] = numNuc;
  /* scan remaining windows */
  lb = 0;
  while(rb < al->n-stepLen+1){
    for(i=0;i<stepLen;i++){
      if(al->nucInd[rb+i])
	numNuc++;
      if(al->nucInd[lb+i])
	numNuc--;
    }
    rb += stepLen;
    lb += stepLen;
    al->winNumNuc[al->numWin++] = numNuc;
  }
}
